@@ -44,17 +44,33 @@ dbus_py_variant_level_get(PyObject *obj)
 {
     PyObject *vl_obj;
     PyObject *key = PyLong_FromVoidPtr(obj);
+    long variant_level;
 
     if (!key) {
-        return 0;
+        return -1;
     }
 
     vl_obj = PyDict_GetItem(_dbus_py_variant_levels, key);
     Py_CLEAR(key);
 
-    if (!vl_obj)
+    if (!vl_obj) {
+        /* PyDict_GetItem() does not set an exception when the key is missing.
+         * In our case, it just means that there was no entry in the variant
+         * dictionary for this object.  Semantically, this is equivalent to a
+         * variant level of 0.
+         */
         return 0;
-    return PyInt_AsLong(vl_obj);
+    }
+    variant_level = NATIVEINT_ASLONG(vl_obj);
+    if (variant_level == -1 && PyErr_Occurred()) {
+        /* variant_level < 0 can never be inserted into the dictionary; see
+         * dbus_py_variant_level_set() below.  The semantics of setting
+         * variant_level < 0 is to delete it from the dictionary.
+         */
+        return -1;
+    }
+    assert(variant_level >= 0);
+    return variant_level;
 }
 
 dbus_bool_t
@@ -76,12 +92,12 @@ dbus_py_variant_level_set(PyObject *obj, long variant_level)
         }
     }
     else {
-        PyObject *vl_obj = PyInt_FromLong(variant_level);
+        PyObject *vl_obj = NATIVEINT_FROMLONG(variant_level);
         if (!vl_obj) {
             Py_CLEAR(key);
             return FALSE;
         }
-        if (PyDict_SetItem (_dbus_py_variant_levels, key, vl_obj) < 0) {
+        if (PyDict_SetItem(_dbus_py_variant_levels, key, vl_obj) < 0) {
             Py_CLEAR(key);
             return FALSE;
         }
@@ -95,7 +111,11 @@ dbus_py_variant_level_getattro(PyObject *obj, PyObject *name)
 {
     PyObject *key, *value;
 
-    if (PyString_Check(name)) {
+#ifdef PY3
+    if (PyUnicode_CompareWithASCIIString(name, "variant_level"))
+        return PyObject_GenericGetAttr(obj, name);
+#else
+    if (PyBytes_Check(name)) {
         Py_INCREF(name);
     }
     else if (PyUnicode_Check(name)) {
@@ -109,13 +129,14 @@ dbus_py_variant_level_getattro(PyObject *obj, PyObject *name)
         return NULL;
     }
 
-    if (strcmp(PyString_AS_STRING(name), "variant_level")) {
+    if (strcmp(PyBytes_AS_STRING(name), "variant_level")) {
         value = PyObject_GenericGetAttr(obj, name);
         Py_CLEAR(name);
         return value;
     }
 
     Py_CLEAR(name);
+#endif  /* PY3 */
 
     key = PyLong_FromVoidPtr(obj);
 
@@ -127,7 +148,7 @@ dbus_py_variant_level_getattro(PyObject *obj, PyObject *name)
     Py_CLEAR(key);
 
     if (!value)
-        return PyInt_FromLong(0);
+        return NATIVEINT_FROMLONG(0);
     Py_INCREF(value);
     return value;
 }
@@ -148,6 +169,7 @@ dbus_py_variant_level_clear(PyObject *self)
     PyErr_Restore(et, ev, etb);
 }
 
+#ifndef PY3
 /* Support code for int subclasses. ================================== */
 
 PyDoc_STRVAR(DBusPythonInt_tp_doc,\
@@ -200,14 +222,14 @@ DBusPythonInt_tp_repr(PyObject *self)
 
     if (!parent_repr) return NULL;
     if (variant_level > 0) {
-        my_repr = PyString_FromFormat("%s(%s, variant_level=%ld)",
-                                      self->ob_type->tp_name,
-                                      PyString_AS_STRING(parent_repr),
-                                      variant_level);
+        my_repr = PyUnicode_FromFormat("%s(%V, variant_level=%ld)",
+                                       Py_TYPE(self)->tp_name,
+                                       REPRV(parent_repr),
+                                       variant_level);
     }
     else {
-        my_repr = PyString_FromFormat("%s(%s)", self->ob_type->tp_name,
-                                      PyString_AS_STRING(parent_repr));
+        my_repr = PyUnicode_FromFormat("%s(%V)", Py_TYPE(self)->tp_name,
+                                       REPRV(parent_repr));
     }
     /* whether my_repr is NULL or not: */
     Py_CLEAR(parent_repr);
@@ -215,8 +237,7 @@ DBusPythonInt_tp_repr(PyObject *self)
 }
 
 PyTypeObject DBusPyIntBase_Type = {
-    PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
-    0,
+    PyVarObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type), 0)
     "_dbus_bindings._IntBase",
     sizeof(DBusPyIntBase),
     0,
@@ -256,6 +277,7 @@ PyTypeObject DBusPyIntBase_Type = {
     DBusPythonInt_tp_new,                   /* tp_new */
     PyObject_Del,                           /* tp_free */
 };
+#endif  /* !PY3 */
 
 /* Support code for float subclasses. ================================ */
 
@@ -312,14 +334,14 @@ DBusPythonFloat_tp_repr(PyObject *self)
 
     if (!parent_repr) return NULL;
     if (variant_level > 0) {
-        my_repr = PyString_FromFormat("%s(%s, variant_level=%ld)",
-                                      self->ob_type->tp_name,
-                                      PyString_AS_STRING(parent_repr),
-                                      variant_level);
+        my_repr = PyUnicode_FromFormat("%s(%V, variant_level=%ld)",
+                                       Py_TYPE(self)->tp_name,
+                                       REPRV(parent_repr),
+                                       variant_level);
     }
     else {
-        my_repr = PyString_FromFormat("%s(%s)", self->ob_type->tp_name,
-                                      PyString_AS_STRING(parent_repr));
+        my_repr = PyUnicode_FromFormat("%s(%V)", Py_TYPE(self)->tp_name,
+                                       REPRV(parent_repr));
     }
     /* whether my_repr is NULL or not: */
     Py_CLEAR(parent_repr);
@@ -327,8 +349,7 @@ DBusPythonFloat_tp_repr(PyObject *self)
 }
 
 PyTypeObject DBusPyFloatBase_Type = {
-    PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
-    0,
+    PyVarObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type), 0)
     "_dbus_bindings._FloatBase",
     sizeof(DBusPyFloatBase),
     0,
@@ -368,6 +389,131 @@ PyTypeObject DBusPyFloatBase_Type = {
     DBusPythonFloat_tp_new,                 /* tp_new */
 };
 
+#ifdef PY3
+/* Support code for bytes subclasses ================================== */
+
+PyDoc_STRVAR(DBusPythonBytes_tp_doc,\
+"Base class for bytes subclasses with a ``variant_level`` attribute.\n"
+"Do not rely on the existence of this class outside dbus-python.\n"
+);
+
+static PyObject *
+DBusPythonBytes_tp_new(PyTypeObject *cls, PyObject *args, PyObject *kwargs)
+{
+    PyObject *self;
+    long variantness = 0;
+    static char *argnames[] = {"variant_level", NULL};
+
+    if (PyTuple_Size(args) > 1) {
+        PyErr_SetString(PyExc_TypeError,
+                        "__new__ takes at most one positional parameter");
+        return NULL;
+    }
+    if (!PyArg_ParseTupleAndKeywords(dbus_py_empty_tuple, kwargs,
+                                     "|l:__new__", argnames,
+                                     &variantness))
+        return NULL;
+
+    if (variantness < 0) {
+        PyErr_SetString(PyExc_ValueError,
+                       "variant_level must be non-negative");
+        return NULL;
+    }
+
+    self = (PyBytes_Type.tp_new)(cls, args, NULL);
+    if (self) {
+        if (!dbus_py_variant_level_set(self, variantness)) {
+            Py_CLEAR(self);
+            return NULL;
+        }
+    }
+    return self;
+}
+
+static PyObject *
+DBusPythonBytes_tp_repr(PyObject *self)
+{
+    PyObject *parent_repr = (PyBytes_Type.tp_repr)(self);
+    PyObject *vl_obj;
+    PyObject *my_repr;
+    long variant_level;
+
+    if (!parent_repr) return NULL;
+    vl_obj = PyObject_GetAttr(self, dbus_py_variant_level_const);
+    if (!vl_obj) {
+        Py_CLEAR(parent_repr);
+        return NULL;
+    }
+    variant_level = NATIVEINT_ASLONG(vl_obj);
+    Py_CLEAR(vl_obj);
+    if (variant_level == -1 && PyErr_Occurred()) {
+        Py_CLEAR(parent_repr);
+        return NULL;
+    }
+    if (variant_level > 0) {
+        my_repr = PyUnicode_FromFormat("%s(%V, variant_level=%ld)",
+                                       Py_TYPE(self)->tp_name,
+                                       REPRV(parent_repr),
+                                       variant_level);
+    }
+    else {
+        my_repr = PyUnicode_FromFormat("%s(%V)", Py_TYPE(self)->tp_name,
+                                       REPRV(parent_repr));
+    }
+    /* whether my_repr is NULL or not: */
+    Py_CLEAR(parent_repr);
+    return my_repr;
+}
+
+static void
+DBusPyBytesBase_tp_dealloc(PyObject *self)
+{
+    dbus_py_variant_level_clear(self);
+    (PyBytes_Type.tp_dealloc)(self);
+}
+
+PyTypeObject DBusPyBytesBase_Type = {
+    PyVarObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type), 0)
+    "_dbus_bindings._BytesBase",
+    0,
+    0,
+    DBusPyBytesBase_tp_dealloc,             /* tp_dealloc */
+    0,                                      /* tp_print */
+    0,                                      /* tp_getattr */
+    0,                                      /* tp_setattr */
+    0,                                      /* tp_compare */
+    DBusPythonBytes_tp_repr,                /* tp_repr */
+    0,                                      /* tp_as_number */
+    0,                                      /* tp_as_sequence */
+    0,                                      /* tp_as_mapping */
+    0,                                      /* tp_hash */
+    0,                                      /* tp_call */
+    0,                                      /* tp_str */
+    dbus_py_variant_level_getattro,         /* tp_getattro */
+    dbus_py_immutable_setattro,             /* tp_setattro */
+    0,                                      /* tp_as_buffer */
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /* tp_flags */
+    DBusPythonBytes_tp_doc,                 /* tp_doc */
+    0,                                      /* tp_traverse */
+    0,                                      /* tp_clear */
+    0,                                      /* tp_richcompare */
+    0,                                      /* tp_weaklistoffset */
+    0,                                      /* tp_iter */
+    0,                                      /* tp_iternext */
+    0,                                      /* tp_methods */
+    0,                                      /* tp_members */
+    0,                                      /* tp_getset */
+    DEFERRED_ADDRESS(&PyBytes_Type),        /* tp_base */
+    0,                                      /* tp_dict */
+    0,                                      /* tp_descr_get */
+    0,                                      /* tp_descr_set */
+    0,                                      /* tp_dictoffset */
+    0,                                      /* tp_init */
+    0,                                      /* tp_alloc */
+    DBusPythonBytes_tp_new,                 /* tp_new */
+};
+#endif  /* PY3 */
+
 /* Support code for str subclasses ================================== */
 
 PyDoc_STRVAR(DBusPythonString_tp_doc,\
@@ -396,7 +542,7 @@ DBusPythonString_tp_new(PyTypeObject *cls, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
-    self = (PyString_Type.tp_new)(cls, args, NULL);
+    self = (NATIVESTR_TYPE.tp_new)(cls, args, NULL);
     if (self) {
         if (!dbus_py_variant_level_set(self, variantness)) {
             Py_CLEAR(self);
@@ -409,7 +555,7 @@ DBusPythonString_tp_new(PyTypeObject *cls, PyObject *args, PyObject *kwargs)
 static PyObject *
 DBusPythonString_tp_repr(PyObject *self)
 {
-    PyObject *parent_repr = (PyString_Type.tp_repr)(self);
+    PyObject *parent_repr = (NATIVESTR_TYPE.tp_repr)(self);
     PyObject *vl_obj;
     PyObject *my_repr;
     long variant_level;
@@ -420,17 +566,22 @@ DBusPythonString_tp_repr(PyObject *self)
         Py_CLEAR(parent_repr);
         return NULL;
     }
-    variant_level = PyInt_AsLong(vl_obj);
+    variant_level = NATIVEINT_ASLONG(vl_obj);
     Py_CLEAR(vl_obj);
+    if (variant_level == -1 && PyErr_Occurred()) {
+        Py_CLEAR(parent_repr);
+        return NULL;
+    }
+
     if (variant_level > 0) {
-        my_repr = PyString_FromFormat("%s(%s, variant_level=%ld)",
-                                      self->ob_type->tp_name,
-                                      PyString_AS_STRING(parent_repr),
-                                      variant_level);
+        my_repr = PyUnicode_FromFormat("%s(%V, variant_level=%ld)",
+                                       Py_TYPE(self)->tp_name,
+                                       REPRV(parent_repr),
+                                       variant_level);
     }
     else {
-        my_repr = PyString_FromFormat("%s(%s)", self->ob_type->tp_name,
-                                      PyString_AS_STRING(parent_repr));
+        my_repr = PyUnicode_FromFormat("%s(%V)", Py_TYPE(self)->tp_name,
+                                       REPRV(parent_repr));
     }
     /* whether my_repr is NULL or not: */
     Py_CLEAR(parent_repr);
@@ -441,12 +592,11 @@ static void
 DBusPyStrBase_tp_dealloc(PyObject *self)
 {
     dbus_py_variant_level_clear(self);
-    (PyString_Type.tp_dealloc)(self);
+    (NATIVESTR_TYPE.tp_dealloc)(self);
 }
 
 PyTypeObject DBusPyStrBase_Type = {
-    PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
-    0,
+    PyVarObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type), 0)
     "_dbus_bindings._StrBase",
     0,
     0,
@@ -476,7 +626,7 @@ PyTypeObject DBusPyStrBase_Type = {
     0,                                      /* tp_methods */
     0,                                      /* tp_members */
     0,                                      /* tp_getset */
-    DEFERRED_ADDRESS(&PyString_Type),       /* tp_base */
+    DEFERRED_ADDRESS(&NATIVESTR_TYPE),      /* tp_base */
     0,                                      /* tp_dict */
     0,                                      /* tp_descr_get */
     0,                                      /* tp_descr_set */
@@ -538,17 +688,22 @@ DBusPythonLong_tp_repr(PyObject *self)
         Py_CLEAR(parent_repr);
         return NULL;
     }
-    variant_level = PyInt_AsLong(vl_obj);
+    variant_level = NATIVEINT_ASLONG(vl_obj);
     Py_CLEAR(vl_obj);
+    if (variant_level < 0 && PyErr_Occurred()) {
+        Py_CLEAR(parent_repr);
+        return NULL;
+    }
+
     if (variant_level) {
-        my_repr = PyString_FromFormat("%s(%s, variant_level=%ld)",
-                                      self->ob_type->tp_name,
-                                      PyString_AS_STRING(parent_repr),
-                                      variant_level);
+        my_repr = PyUnicode_FromFormat("%s(%V, variant_level=%ld)",
+                                       Py_TYPE(self)->tp_name,
+                                       REPRV(parent_repr),
+                                       variant_level);
     }
     else {
-        my_repr = PyString_FromFormat("%s(%s)", self->ob_type->tp_name,
-                                      PyString_AS_STRING(parent_repr));
+        my_repr = PyUnicode_FromFormat("%s(%V)", Py_TYPE(self)->tp_name,
+                                       REPRV(parent_repr));
     }
     /* whether my_repr is NULL or not: */
     Py_CLEAR(parent_repr);
@@ -563,8 +718,7 @@ DBusPyLongBase_tp_dealloc(PyObject *self)
 }
 
 PyTypeObject DBusPyLongBase_Type = {
-    PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
-    0,
+    PyVarObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type), 0)
     "_dbus_bindings._LongBase",
     0,
     0,
@@ -608,26 +762,41 @@ PyObject *dbus_py_variant_level_const = NULL;
 PyObject *dbus_py_signature_const = NULL;
 PyObject *dbus_py__dbus_object_path__const = NULL;
 
+#ifdef PY3
+#define INTERN (PyUnicode_InternFromString)
+#else
+/* Neither Python 2.6 nor 2.7 define the expected PyBytes_InternFromString
+ * alias in bytesobject.h.
+ */
+#define INTERN (PyString_InternFromString)
+#endif
+
 dbus_bool_t
 dbus_py_init_abstract(void)
 {
     _dbus_py_variant_levels = PyDict_New();
     if (!_dbus_py_variant_levels) return 0;
 
-    dbus_py__dbus_object_path__const = PyString_InternFromString("__dbus_object_path__");
+    dbus_py__dbus_object_path__const = INTERN("__dbus_object_path__");
     if (!dbus_py__dbus_object_path__const) return 0;
 
-    dbus_py_variant_level_const = PyString_InternFromString("variant_level");
+    dbus_py_variant_level_const = INTERN("variant_level");
     if (!dbus_py_variant_level_const) return 0;
 
-    dbus_py_signature_const = PyString_InternFromString("signature");
+    dbus_py_signature_const = INTERN("signature");
     if (!dbus_py_signature_const) return 0;
 
+#ifdef PY3
+    DBusPyBytesBase_Type.tp_base = &PyBytes_Type;
+    if (PyType_Ready(&DBusPyBytesBase_Type) < 0) return 0;
+    DBusPyBytesBase_Type.tp_print = NULL;
+#else
     DBusPyIntBase_Type.tp_base = &PyInt_Type;
     if (PyType_Ready(&DBusPyIntBase_Type) < 0) return 0;
     /* disable the tp_print copied from PyInt_Type, so tp_repr gets called as
     desired */
     DBusPyIntBase_Type.tp_print = NULL;
+#endif
 
     DBusPyFloatBase_Type.tp_base = &PyFloat_Type;
     if (PyType_Ready(&DBusPyFloatBase_Type) < 0) return 0;
@@ -637,7 +806,7 @@ dbus_py_init_abstract(void)
     if (PyType_Ready(&DBusPyLongBase_Type) < 0) return 0;
     DBusPyLongBase_Type.tp_print = NULL;
 
-    DBusPyStrBase_Type.tp_base = &PyString_Type;
+    DBusPyStrBase_Type.tp_base = &NATIVESTR_TYPE;
     if (PyType_Ready(&DBusPyStrBase_Type) < 0) return 0;
     DBusPyStrBase_Type.tp_print = NULL;
 
@@ -648,12 +817,18 @@ dbus_bool_t
 dbus_py_insert_abstract_types(PyObject *this_module)
 {
     /* PyModule_AddObject steals a ref */
+#ifdef PY3
+    Py_INCREF(&DBusPyBytesBase_Type);
+    if (PyModule_AddObject(this_module, "_BytesBase",
+                           (PyObject *)&DBusPyBytesBase_Type) < 0) return 0;
+#else
     Py_INCREF(&DBusPyIntBase_Type);
+    if (PyModule_AddObject(this_module, "_IntBase",
+                           (PyObject *)&DBusPyIntBase_Type) < 0) return 0;
+#endif
     Py_INCREF(&DBusPyLongBase_Type);
     Py_INCREF(&DBusPyStrBase_Type);
     Py_INCREF(&DBusPyFloatBase_Type);
-    if (PyModule_AddObject(this_module, "_IntBase",
-                           (PyObject *)&DBusPyIntBase_Type) < 0) return 0;
     if (PyModule_AddObject(this_module, "_LongBase",
                            (PyObject *)&DBusPyLongBase_Type) < 0) return 0;
     if (PyModule_AddObject(this_module, "_StrBase",
