@@ -42,9 +42,11 @@ char dbus_py_Message_get_args_list__doc__[] = (
 "       it's off by default for consistency.\n"
 "\n"
 "       If false (default), convert them into a dbus.Array of Bytes.\n"
+#ifndef PY3
 "   `utf8_strings` : bool\n"
 "       If true, return D-Bus strings as Python 8-bit strings (of UTF-8).\n"
 "       If false (default), return D-Bus strings as Python unicode objects.\n"
+#endif
 "\n"
 "Most of the type mappings should be fairly obvious:\n"
 "\n"
@@ -70,7 +72,9 @@ char dbus_py_Message_get_args_list__doc__[] = (
 
 typedef struct {
     int byte_arrays;
+#ifndef PY3
     int utf8_strings;
+#endif
 } Message_get_args_options;
 
 static PyObject *_message_iter_get_pyobject(DBusMessageIter *iter,
@@ -94,7 +98,7 @@ _message_iter_append_all_to_list(DBusMessageIter *iter, PyObject *list,
 #ifdef USING_DBG
         fprintf(stderr, "DBG/%ld: appending to list: %p == ", (long)getpid(), item);
         PyObject_Print(item, stderr, 0);
-        fprintf(stderr, " of type %p\n", item->ob_type);
+        fprintf(stderr, " of type %p\n", Py_TYPE(item));
 #endif
         ret = PyList_Append(list, item);
         Py_CLEAR(item);
@@ -213,7 +217,7 @@ _message_iter_get_pyobject(DBusMessageIter *iter,
     if (variant_level > 0 && type != DBUS_TYPE_VARIANT) {
         PyObject *variant_level_int;
 
-        variant_level_int = PyInt_FromLong(variant_level);
+        variant_level_int = NATIVEINT_FROMLONG(variant_level);
         if (!variant_level_int) {
             return NULL;
         }
@@ -235,9 +239,12 @@ _message_iter_get_pyobject(DBusMessageIter *iter,
      */
 
     switch (type) {
+        PyObject *unicode;
+
         case DBUS_TYPE_STRING:
             DBG("%s", "found a string");
             dbus_message_iter_get_basic(iter, &u.s);
+#ifndef PY3
             if (opts->utf8_strings) {
                 args = Py_BuildValue("(s)", u.s);
                 if (!args) break;
@@ -245,15 +252,20 @@ _message_iter_get_pyobject(DBusMessageIter *iter,
                                     args, kwargs);
             }
             else {
-                args = Py_BuildValue("(N)", PyUnicode_DecodeUTF8(u.s,
-                                                                 strlen(u.s),
-                                                                 NULL));
+#endif
+                unicode = PyUnicode_DecodeUTF8(u.s, strlen(u.s), NULL);
+                if (!unicode) {
+                    break;
+                }
+                args = Py_BuildValue("(N)", unicode);
                 if (!args) {
                     break;
                 }
                 ret = PyObject_Call((PyObject *)&DBusPyString_Type,
                                     args, kwargs);
+#ifndef PY3
             }
+#endif
             break;
 
         case DBUS_TYPE_SIGNATURE:
@@ -408,7 +420,11 @@ _message_iter_get_pyobject(DBusMessageIter *iter,
                      * for an empty byte-blob... */
                     u.s = "";
                 }
+#ifdef PY3
+                args = Py_BuildValue("(y#)", u.s, (Py_ssize_t)n);
+#else
                 args = Py_BuildValue("(s#)", u.s, (Py_ssize_t)n);
+#endif
                 if (!args) break;
                 ret = PyObject_Call((PyObject *)&DBusPyByteArray_Type,
                                     args, kwargs);
@@ -492,8 +508,13 @@ _message_iter_get_pyobject(DBusMessageIter *iter,
 PyObject *
 dbus_py_Message_get_args_list(Message *self, PyObject *args, PyObject *kwargs)
 {
+#ifdef PY3
+    Message_get_args_options opts = { 0 };
+    static char *argnames[] = { "byte_arrays", NULL };
+#else
     Message_get_args_options opts = { 0, 0 };
     static char *argnames[] = { "byte_arrays", "utf8_strings", NULL };
+#endif
     PyObject *list;
     DBusMessageIter iter;
 
@@ -513,10 +534,16 @@ dbus_py_Message_get_args_list(Message *self, PyObject *args, PyObject *kwargs)
                         "arguments");
         return NULL;
     }
+#ifdef PY3
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|i:get_args_list",
+                                     argnames,
+                                     &(opts.byte_arrays))) return NULL;
+#else
     if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|ii:get_args_list",
                                      argnames,
                                      &(opts.byte_arrays),
                                      &(opts.utf8_strings))) return NULL;
+#endif
     if (!self->msg) return DBusPy_RaiseUnusableMessage();
 
     list = PyList_New(0);
